#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import textwrap
from unittest import mock

from oslotest import base

from oslo_config import cfg
from oslo_config import sphinxext
from oslo_config import types


class FormatGroupTest(base.BaseTestCase):
    def test_none_in_default(self):
        # option with None group placed in DEFAULT
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.StrOpt(
                            'opt_name',
                            help='this appears in the default group',
                        ),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: string
          :Default: ``<None>``

          this appears in the default group
        ''').lstrip(),
            results,
        )

    def test_with_default_value(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.StrOpt(
                            'opt_name',
                            default='this is the default',
                            help='this appears in the default group',
                        ),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: string
          :Default: ``this is the default``

          this appears in the default group
        ''').lstrip(),
            results,
        )

    def test_with_custom_type(self):
        TEST_TYPE_NAME = 'test type'

        class TestType(types.ConfigType):
            def __init__(self):
                super().__init__(type_name=TEST_TYPE_NAME)

            def __call__(self, value):
                return str(value)

            def _formatter(self, value):
                return self.quote_trailing_and_leading_space(value)

        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.Opt(
                            'custom_type_opt_name',
                            type=TestType(),
                            help='this is a custom option type',
                        ),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: custom_type_opt_name

          :Type: test type
          :Default: ``<None>``

          this is a custom option type
        ''').lstrip(),
            results,
        )

    def test_with_min(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.IntOpt('opt_name', min=1),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: integer
          :Default: ``<None>``
          :Minimum Value: 1
        ''').lstrip(),
            results,
        )

    def test_with_min_0(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.IntOpt('opt_name', min=0),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: integer
          :Default: ``<None>``
          :Minimum Value: 0
        ''').lstrip(),
            results,
        )

    def test_with_max(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.IntOpt('opt_name', max=1),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: integer
          :Default: ``<None>``
          :Maximum Value: 1
        ''').lstrip(),
            results,
        )

    def test_with_max_0(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.IntOpt('opt_name', max=0),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: integer
          :Default: ``<None>``
          :Maximum Value: 0
        ''').lstrip(),
            results,
        )

    def test_with_choices(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.StrOpt(
                            'opt_name', choices=['a', 'b', 'c', None, '']
                        ),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: string
          :Default: ``<None>``
          :Valid Values: a, b, c, <None>, ''
        ''').lstrip(),
            results,
        )

    def test_with_choices_with_descriptions(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.StrOpt(
                            'opt_name',
                            choices=[
                                ('a', 'a is the best'),
                                ('b', 'Actually, may-b I am better'),
                                ('c', 'c, I am clearly the greatest'),
                                (None, 'I am having none of this'),
                                ('', ''),
                            ],
                        ),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: string
          :Default: ``<None>``
          :Valid Values: a, b, c, <None>, ''

          .. rubric:: Possible values

          a
            a is the best

          b
            Actually, may-b I am better

          c
            c, I am clearly the greatest

          <None>
            I am having none of this

          ''
            <No description provided>
        ''').lstrip(),
            results,
        )

    def test_group_obj_without_help(self):
        # option with None group placed in DEFAULT
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name='group',
                    group_obj=cfg.OptGroup('group'),
                    opt_list=[cfg.StrOpt('opt_name')],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: group

        .. oslo.config:option:: opt_name

          :Type: string
          :Default: ``<None>``
        ''').lstrip(),
            results,
        )

    def test_group_obj_with_help(self):
        # option with None group placed in DEFAULT
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name='group',
                    group_obj=cfg.OptGroup('group', help='group help'),
                    opt_list=[cfg.StrOpt('opt_name')],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: group

          group help

        .. oslo.config:option:: opt_name

          :Type: string
          :Default: ``<None>``
        ''').lstrip(),
            results,
        )

    def test_deprecated_opts_without_deprecated_group(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.StrOpt(
                            'opt_name',
                            deprecated_name='deprecated_name',
                        )
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: string
          :Default: ``<None>``

          .. list-table:: Deprecated Variations
             :header-rows: 1

             - * Group
               * Name
             - * DEFAULT
               * deprecated_name
        ''').lstrip(),
            results,
        )

    def test_deprecated_opts_with_deprecated_group(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.StrOpt(
                            'opt_name',
                            deprecated_name='deprecated_name',
                            deprecated_group='deprecated_group',
                        )
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: string
          :Default: ``<None>``

          .. list-table:: Deprecated Variations
             :header-rows: 1

             - * Group
               * Name
             - * deprecated_group
               * deprecated_name
        ''').lstrip(),
            results,
        )

    def test_deprecated_for_removal(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.StrOpt(
                            'opt_name',
                            deprecated_for_removal=True,
                            deprecated_reason='because I said so',
                            deprecated_since='13.0',
                        )
                    ],
                )
            )
        )
        self.assertIn('.. warning::', results)
        self.assertIn('because I said so', results)
        self.assertIn('since 13.0', results)

    def test_mutable(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.IntOpt('opt_name', mutable=True),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: integer
          :Default: ``<None>``
          :Mutable: This option can be changed without restarting.
        ''').lstrip(),
            results,
        )

    def test_not_mutable(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.IntOpt('opt_name', mutable=False),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: integer
          :Default: ``<None>``
        ''').lstrip(),
            results,
        )

    def test_advanced(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.StrOpt('opt_name', advanced=True),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: string
          :Default: ``<None>``
          :Advanced Option: Intended for advanced users and not used
              by the majority of users, and might have a significant
              effect on stability and/or performance.
        ''').lstrip(),
            results,
        )

    def test_not_advanced(self):
        results = '\n'.join(
            list(
                sphinxext._format_group_opts(
                    namespace=None,
                    group_name=None,
                    group_obj=None,
                    opt_list=[
                        cfg.StrOpt('opt_name', advanced=False),
                    ],
                )
            )
        )
        self.assertEqual(
            textwrap.dedent('''
        .. oslo.config:group:: DEFAULT

        .. oslo.config:option:: opt_name

          :Type: string
          :Default: ``<None>``
        ''').lstrip(),
            results,
        )


class FormatOptionHelpTest(base.BaseTestCase):
    @mock.patch('oslo_config.generator._list_opts')
    @mock.patch('oslo_config.sphinxext._format_group_opts')
    def test_split_namespaces(self, _format_group_opts, _list_opts):
        _list_opts.return_value = [
            ('namespace1', [(None, ['opt1'])]),
            ('namespace2', [(None, ['opt2'])]),
        ]
        list(
            sphinxext._format_option_help(
                namespaces=['namespace1', 'namespace2'], split_namespaces=True
            )
        )
        _format_group_opts.assert_any_call(
            namespace='namespace1',
            group_name='DEFAULT',
            group_obj=None,
            opt_list=['opt1'],
        )
        _format_group_opts.assert_any_call(
            namespace='namespace2',
            group_name='DEFAULT',
            group_obj=None,
            opt_list=['opt2'],
        )

    @mock.patch('oslo_config.generator._list_opts')
    @mock.patch('oslo_config.sphinxext._format_group_opts')
    def test_dont_split_namespaces(self, _format_group_opts, _list_opts):
        _list_opts.return_value = [
            ('namespace1', [(None, ['opt1'])]),
            ('namespace2', [(None, ['opt2'])]),
        ]
        list(
            sphinxext._format_option_help(
                namespaces=['namespace1', 'namespace2'], split_namespaces=False
            )
        )
        _format_group_opts.assert_called_once_with(
            namespace=None,
            group_name='DEFAULT',
            group_obj=None,
            opt_list=['opt1', 'opt2'],
        )

    @mock.patch('oslo_config.generator._list_opts')
    @mock.patch('oslo_config.sphinxext._format_group_opts')
    def test_dont_split_namespaces_with_group(
        self, _format_group_opts, _list_opts
    ):
        grp_obj = cfg.OptGroup('grp1')
        _list_opts.return_value = [
            ('namespace1', [(grp_obj, ['opt1'])]),
            ('namespace2', [('grp1', ['opt2'])]),
        ]
        list(
            sphinxext._format_option_help(
                namespaces=['namespace1', 'namespace2'], split_namespaces=False
            )
        )
        _format_group_opts.assert_any_call(
            namespace=None,
            group_name='grp1',
            group_obj=grp_obj,
            opt_list=['opt1', 'opt2'],
        )

    @mock.patch('oslo_config.generator._list_opts')
    @mock.patch('oslo_config.sphinxext._format_group_opts')
    def test_split_namespaces_with_group(self, _format_group_opts, _list_opts):
        grp_obj = cfg.OptGroup('grp1')
        _list_opts.return_value = [
            ('namespace1', [(grp_obj, ['opt1'])]),
            ('namespace2', [('grp1', ['opt2'])]),
        ]
        list(
            sphinxext._format_option_help(
                namespaces=['namespace1', 'namespace2'], split_namespaces=True
            )
        )
        print(_format_group_opts.call_args_list)
        _format_group_opts.assert_any_call(
            namespace='namespace1',
            group_name='grp1',
            group_obj=grp_obj,
            opt_list=['opt1'],
        )
        _format_group_opts.assert_any_call(
            namespace='namespace2',
            group_name='grp1',
            group_obj=None,
            opt_list=['opt2'],
        )



#
# Copyright 2013 Mirantis, Inc.
# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from typing import Any

import fixtures

from oslo_config import cfg


class Config(fixtures.Fixture):
    """Allows overriding configuration settings for the test.

    `conf` will be reset on cleanup.

    """

    conf: cfg.ConfigOpts
    _registered_config_opts: dict[str | None, set[cfg.Opt]]
    _default_config_files: list[str] | None
    _default_config_dirs: list[str] | None

    def __init__(self, conf: cfg.ConfigOpts = cfg.CONF) -> None:
        self.conf = conf

    def setUp(self) -> None:
        super().setUp()
        # NOTE(morganfainberg): unregister must be added to cleanup before
        # reset is because cleanup works in reverse order of registered items,
        # and a reset must occur before unregistering options can occur.
        self.addCleanup(self._reset_default_config_files)
        self.addCleanup(self._reset_default_config_dirs)
        self.addCleanup(self._unregister_config_opts)
        self.addCleanup(self.conf.reset)
        self._registered_config_opts = {}

        # Grab an old copy of the default config files/dirs - if it exists -
        # for subsequent cleanup.
        if hasattr(self.conf, 'default_config_files'):
            self._default_config_files = self.conf.default_config_files
        else:
            self._default_config_files = None

        if hasattr(self.conf, 'default_config_dirs'):
            self._default_config_dirs = self.conf.default_config_dirs
        else:
            self._default_config_dirs = None

    def config(self, **kw: Any) -> None:
        """Override configuration values.

        The keyword arguments are the names of configuration options to
        override and their values.

        If a `group` argument is supplied, the overrides are applied to
        the specified configuration option group, otherwise the overrides
        are applied to the ``default`` group.

        """

        group = kw.pop('group', None)
        for k, v in kw.items():
            self.conf.set_override(k, v, group)

    def _unregister_config_opts(self) -> None:
        for group in self._registered_config_opts:
            self.conf.unregister_opts(
                self._registered_config_opts[group], group=group
            )

    def _reset_default_config_files(self) -> None:
        if not hasattr(self.conf, 'default_config_files'):
            return

        if self._default_config_files is not None:
            self.conf.default_config_files = self._default_config_files
        else:
            # Delete, because we could conceivably begin with the property
            # being unset.
            del self.conf.default_config_files

    def _reset_default_config_dirs(self) -> None:
        if not hasattr(self.conf, 'default_config_dirs'):
            return

        if self._default_config_dirs is not None:
            self.conf.default_config_dirs = self._default_config_dirs
        else:
            # Delete, because we could conceivably begin with the property
            # being unset.
            del self.conf.default_config_dirs

    def register_opt(self, opt: cfg.Opt, group: str | None = None) -> None:
        """Register a single option for the test run.

        Options registered in this manner will automatically be unregistered
        during cleanup.

        If a `group` argument is supplied, it will register the new option
        to that group, otherwise the option is registered to the ``default``
        group.
        """
        self.conf.register_opt(opt, group=group)
        self._registered_config_opts.setdefault(group, set()).add(opt)

    def register_opts(
        self, opts: list[cfg.Opt], group: str | None = None
    ) -> None:
        """Register multiple options for the test run.

        This works in the same manner as register_opt() but takes a list of
        options as the first argument. All arguments will be registered to the
        same group if the ``group`` argument is supplied, otherwise all options
        will be registered to the ``default`` group.
        """
        for opt in opts:
            self.register_opt(opt, group=group)

    def register_cli_opt(self, opt: cfg.Opt, group: str | None = None) -> None:
        """Register a single CLI option for the test run.

        Options registered in this manner will automatically be unregistered
        during cleanup.

        If a `group` argument is supplied, it will register the new option
        to that group, otherwise the option is registered to the ``default``
        group.

        CLI options must be registered before the command line and config files
        are parsed. This is to ensure that all CLI options are shown in --help
        and option validation works as expected.
        """
        self.conf.register_cli_opt(opt, group=group)
        self._registered_config_opts.setdefault(group, set()).add(opt)

    def register_cli_opts(
        self, opts: list[cfg.Opt], group: str | None = None
    ) -> None:
        """Register multiple CLI options for the test run.

        This works in the same manner as register_opt() but takes a list of
        options as the first argument. All arguments will be registered to the
        same group if the ``group`` argument is supplied, otherwise all options
        will be registered to the ``default`` group.

        CLI options must be registered before the command line and config files
        are parsed. This is to ensure that all CLI options are shown in --help
        and option validation works as expected.
        """
        for opt in opts:
            self.register_cli_opt(opt, group=group)

    def load_raw_values(self, group: str | None = None, **kwargs: Any) -> None:
        """Load raw values into the configuration without registering them.

        This method adds a series of parameters into the current config
        instance, as if they had been loaded by a ConfigParser. This method
        does not require that you register the configuration options first,
        however the values loaded will not be accessible until you do.
        """

        # Make sure the namespace exists for our tests.
        if not self.conf._namespace:
            self.conf.__call__(args=[])

        # Default out the group name
        group = 'DEFAULT' if not group else group

        raw_config: dict[str, dict[str, list[str]]] = {}
        raw_config[group] = {}
        for key, value in kwargs.items():
            # Parsed values are an array of raw strings.
            raw_config[group][key] = [str(value)]

        assert self.conf._namespace is not None
        self.conf._namespace._add_parsed_config_file(
            '<memory>', raw_config, raw_config
        )

    def set_config_files(self, config_files: list[str]) -> None:
        """Specify a list of config files to read.

        This method allows you to predefine the list of configuration files
        that are loaded by oslo_config. It will ensure that your tests do not
        attempt to autodetect, and accidentally pick up config files from
        locally installed services.
        """
        if not isinstance(config_files, list):
            raise AttributeError("Please pass a list() to set_config_files()")

        # Make sure the namespace exists for our tests.
        if not self.conf._namespace:
            self.conf.__call__(args=[])

        self.conf.default_config_files = config_files
        self.conf.reload_config_files()

    def set_config_dirs(self, config_dirs: list[str]) -> None:
        """Specify a list of config dirs to read.

        This method allows you to predefine the list of configuration dirs
        that are loaded by oslo_config. It will ensure that your tests do not
        attempt to autodetect, and accidentally pick up config files from
        locally installed services.
        """
        if not isinstance(config_dirs, list):
            raise AttributeError("Please pass a list() to set_config_dirs()")

        # Make sure the namespace exists for our tests.
        if not self.conf._namespace:
            self.conf([])

        self.conf.default_config_dirs = config_dirs
        self.conf.reload_config_files()

    def set_default(
        self, name: str, default: Any, group: str | None = None
    ) -> None:
        """Set a default value for an option.

        This method is not necessarily meant to be invoked
        directly. It is here to allow the set_defaults() functions in
        various Oslo libraries to work with a Config fixture instead
        of a ConfigOpts instance.

        Use it like::

            class MyTest(testtools.TestCase):
                def setUp(self):
                    super(MyTest, self).setUp()
                    self.conf = self.useFixture(fixture.Config())

                def test_something(self):
                    some_library.set_defaults(self.conf, name='value')
                    some_library.do_something_exciting()

        """
        self.conf.set_default(name, default, group)
        self.addCleanup(self.conf.clear_default, name, group)



#!/usr/bin/env python3
# Copyright 2018 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Configuration Validator

Uses the sample config generator configuration file to retrieve a list of all
the available options in a project, then compares it to what is configured in
the provided file.  If there are any options set that are not defined in the
project then it returns those errors.
"""

import importlib.metadata
import logging
import re
import sys
from typing import Any, cast

import yaml

from oslo_config import cfg
from oslo_config import generator

VALIDATE_DEFAULTS_EXCLUSIONS = [
    '.*_ur(i|l)',
    '.*connection',
    'password',
    'username',
    'my_ip',
    'host(name)?',
    'glance_api_servers',
    'osapi_volume_listen',
    'osapi_compute_listen',
]

_validator_opts = [
    cfg.MultiStrOpt(
        'namespace',
        help='Option namespace under "oslo.config.opts" in which to query '
        'for options.',
    ),
    cfg.StrOpt('input-file', required=True, help='Config file to validate.'),
    cfg.StrOpt(
        'opt-data',
        help='Path to a YAML file containing definitions of options, as '
        'output by the config generator.',
    ),
    cfg.BoolOpt(
        'check-defaults',
        default=False,
        help='Report differences between the sample values and current '
        'values.',
    ),
    cfg.ListOpt(
        'exclude-options',
        default=VALIDATE_DEFAULTS_EXCLUSIONS,
        help='Exclude options matching these patterns when comparing '
        'the current and sample configurations.',
    ),
    cfg.BoolOpt(
        'fatal-warnings',
        default=False,
        help='Report failure if any warnings are found.',
    ),
    cfg.MultiStrOpt(
        'exclude-group',
        default=[],
        help='Groups that should not be validated if they are present in the '
        'specified input-file. This may be necessary for dynamically '
        'named groups which do not appear in the sample config data.',
    ),
]


LOG = logging.getLogger(__name__)


KNOWN_BAD_GROUPS = ['keystone_authtoken']


def _register_cli_opts(conf: cfg.ConfigOpts) -> None:
    """Register the formatter's CLI options with a ConfigOpts instance.

    Note, this must be done before the ConfigOpts instance is called to parse
    the configuration.

    :param conf: a ConfigOpts instance
    :raises: DuplicateOptError, ArgsAlreadyParsedError
    """
    conf.register_cli_opts(_validator_opts)


def _validate_deprecated_opt(
    group: str, option: str, opt_data: dict[str, Any]
) -> bool:
    if group not in opt_data['deprecated_options']:
        return False
    name_data = [o['name'] for o in opt_data['deprecated_options'][group]]
    name_data += [o.get('dest') for o in opt_data['deprecated_options'][group]]
    return option in name_data


def _validate_defaults(
    sections: dict[str, Any], opt_data: dict[str, Any], conf: cfg.ConfigOpts
) -> bool:
    """Compares the current and sample configuration and reports differences

    :param section: ConfigParser instance
    :param opt_data: machine readable data from the generator instance
    :param conf: ConfigOpts instance
    :returns: boolean wether or not warnings were reported
    """
    warnings = False
    # Generating regex objects from ListOpt
    exclusion_regexes = []
    for pattern in conf.exclude_options:
        exclusion_regexes.append(re.compile(pattern))
    for group, opts in opt_data['options'].items():
        if group in conf.exclude_group:
            continue
        if group not in sections:
            LOG.warning(
                'Group %s from the sample config is not defined in input-file',
                group,
            )
            continue
        for opt in opts['opts']:
            # We need to convert the defaults into a list to find
            # intersections. defaults are only a list if they can
            # be defined multiple times, but configparser only
            # returns list
            if not isinstance(opt['default'], list):
                defaults = [str(opt['default'])]
            else:
                defaults = opt['default']

            # Apparently, there's multiple naming conventions for
            # options, 'name' is mostly with hyphens, and 'dest'
            # is represented with underscores.
            opt_names = {opt['name'], opt.get('dest')}
            if not opt_names.intersection(sections[group]):
                continue
            try:
                value = sections[group][opt['name']]
                keyname = opt['name']
            except KeyError:
                value = sections[group][opt.get('dest')]
                keyname = opt.get('dest')

            if any(rex.fullmatch(keyname) for rex in exclusion_regexes):
                LOG.info(
                    '%s/%s Ignoring option because it is part of the excluded '
                    'patterns. This can be changed with the --exclude-options '
                    'argument',
                    group,
                    keyname,
                )
                continue

            if len(value) > 1:
                LOG.info('%s/%s defined %s times', group, keyname, len(value))
            if not opt['default']:
                LOG.warning(
                    '%s/%s sample value is empty but input-file has %s',
                    group,
                    keyname,
                    ", ".join(value),
                )
                warnings = True
            elif not frozenset(defaults).intersection(value):
                LOG.warning(
                    '%s/%s sample value %s is not in %s',
                    group,
                    keyname,
                    defaults,
                    value,
                )
                warnings = True
    return warnings


def _validate_opt(group: str, option: str, opt_data: dict[str, Any]) -> bool:
    if group not in opt_data['options']:
        return False
    name_data = [o['name'] for o in opt_data['options'][group]['opts']]
    name_data += [o.get('dest') for o in opt_data['options'][group]['opts']]
    return option in name_data


def load_opt_data(conf: cfg.ConfigOpts) -> dict[str, Any]:
    with open(conf.opt_data) as f:
        return cast(dict[str, Any], yaml.safe_load(f))


def _validate(conf: cfg.ConfigOpts) -> int:
    conf.register_opts(_validator_opts)
    if conf.namespace:
        groups = generator._get_groups(generator._list_opts(conf.namespace))
        opt_data = generator._generate_machine_readable_data(groups, conf)
    elif conf.opt_data:
        opt_data = load_opt_data(conf)
    else:
        # TODO(bnemec): Implement this logic with group?
        raise RuntimeError('Neither namespace nor opt-data provided.')
    sections: dict[str, Any] = {}
    parser = cfg.ConfigParser(conf.input_file, sections)
    parser.parse()
    warnings = False
    errors = False
    if conf.check_defaults:
        warnings = _validate_defaults(sections, opt_data, conf)
    for section, options in sections.items():
        if section in conf.exclude_group:
            continue
        for option in options:
            if _validate_deprecated_opt(section, option, opt_data):
                LOG.warning('Deprecated opt %s/%s found', section, option)
                warnings = True
            elif not _validate_opt(section, option, opt_data):
                if section in KNOWN_BAD_GROUPS:
                    LOG.info(
                        'Ignoring missing option "%s" from group '
                        '"%s" because the group is known to have '
                        'incomplete sample config data and thus '
                        'cannot be validated properly.',
                        option,
                        section,
                    )
                    continue
                LOG.error(
                    '%s/%s is not part of the sample config', section, option
                )
                errors = True
    if errors or (warnings and conf.fatal_warnings):
        return 1
    return 0


def main() -> int:
    """The main function of oslo-config-validator."""
    version = importlib.metadata.version('oslo.config')
    logging.basicConfig(level=logging.INFO)
    conf = cfg.ConfigOpts()
    _register_cli_opts(conf)
    try:
        conf(sys.argv[1:], version=version)
    except cfg.RequiredOptError:
        conf.print_help()
        if not sys.argv[1:]:
            raise SystemExit(1)
        raise
    return _validate(conf)


if __name__ == '__main__':
    sys.exit(main())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Generator, Iterable, Set
from typing import Any, cast

from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils.statemachine import StringList
import oslo_i18n
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.builders import Builder
from sphinx.directives import ObjectDescription
from sphinx.domains import Domain
from sphinx.domains import ObjType
from sphinx.environment import BuildEnvironment
from sphinx.roles import XRefRole
from sphinx.util import logging
from sphinx.util.nodes import make_refnode
from sphinx.util.nodes import nested_parse_with_titles

from oslo_config import cfg
from oslo_config import generator

LOG = logging.getLogger(__name__)


def _list_table(
    headers: list[str],
    data: Iterable[tuple[str, str]],
    title: str = '',
    columns: list[int] | None = None,
) -> Generator[str, None, None]:
    """Build a list-table directive.

    :param add: Function to add one row to output.
    :param headers: List of header values.
    :param data: Iterable of row data, yielding lists or tuples with rows.
    """
    yield f'.. list-table:: {title}'
    yield '   :header-rows: 1'
    if columns:
        yield '   :widths: {}'.format(','.join(str(c) for c in columns))
    yield ''
    yield f'   - * {headers[0]}'
    for h in headers[1:]:
        yield f'     * {h}'
    for row in data:
        yield f'   - * {row[0]}'
        for r in row[1:]:
            yield f'     * {r}'


def _indent(text: str, n: int = 2) -> str:
    padding = ' ' * n
    # we don't want to indent blank lines so just output them as-is
    return '\n'.join(padding + x if x else '' for x in text.splitlines())


def _make_anchor_target(group_name: str, option_name: str) -> str:
    # We need to ensure this is unique across entire documentation
    # http://www.sphinx-doc.org/en/stable/markup/inline.html#ref-role
    target = f'{cfg._normalize_group_name(group_name)}.{option_name.lower()}'
    return target


_TYPE_DESCRIPTIONS = {
    cfg.StrOpt: 'string',
    cfg.BoolOpt: 'boolean',
    cfg.IntOpt: 'integer',
    cfg.FloatOpt: 'floating point',
    cfg.ListOpt: 'list',
    cfg.DictOpt: 'dict',
    cfg.MultiStrOpt: 'multi-valued',
    cfg.IPOpt: 'ip address',
    cfg.PortOpt: 'port number',
    cfg.HostnameOpt: 'hostname',
    cfg.URIOpt: 'URI',
    cfg.HostAddressOpt: 'host address',
    cfg._ConfigFileOpt: 'list of filenames',
    cfg._ConfigDirOpt: 'list of directory names',
}


_UNKNOWN_TYPE = 'unknown type'


def _get_choice_text(choice: Any) -> str:
    if choice is None:
        return '<None>'
    elif choice == '':
        return "''"
    return str(choice)


def _format_opt(opt: Any, group_name: str) -> Generator[str, None, None]:
    opt_type = _TYPE_DESCRIPTIONS.get(type(opt), _UNKNOWN_TYPE)

    # NOTE(clif): If we don't have a type description handy, fall back to
    # trying to retrieve the type from opt.type if available.
    if opt_type == _UNKNOWN_TYPE and hasattr(opt, 'type'):
        opt_type = getattr(opt.type, 'type_name', _UNKNOWN_TYPE)

    yield f'.. oslo.config:option:: {opt.dest}'
    yield ''
    yield _indent(f':Type: {opt_type}')
    for default in generator._format_defaults(opt):
        if default:
            yield _indent(f':Default: ``{default}``')
        else:
            yield _indent(f':Default: ``{default!r}``')
    if getattr(opt.type, 'min', None) is not None:
        yield _indent(f':Minimum Value: {opt.type.min}')
    if getattr(opt.type, 'max', None) is not None:
        yield _indent(f':Maximum Value: {opt.type.max}')
    if getattr(opt.type, 'choices', None):
        choices_text = ', '.join(
            [_get_choice_text(choice) for choice in opt.type.choices]
        )
        yield _indent(f':Valid Values: {choices_text}')
    try:
        if opt.mutable:
            yield _indent(
                ':Mutable: This option can be changed without restarting.'
            )
    except AttributeError as err:
        # NOTE(dhellmann): keystoneauth defines its own Opt class,
        # and neutron (at least) returns instances of those
        # classes instead of oslo_config Opt instances. The new
        # mutable attribute is the first property where the API
        # isn't supported in the external class, so we can use
        # this failure to emit a warning. See
        # https://bugs.launchpad.net/keystoneauth/+bug/1548433 for
        # more details.
        import warnings

        if not isinstance(cfg.Opt, opt):
            warnings.warn(
                f'Incompatible option class for {opt.dest} '
                f'({opt.__class__!r}): {err}',
            )
        else:
            warnings.warn(
                f'Failed to fully format sample for {opt.dest}: {err}'
            )
    if opt.advanced:
        yield _indent(
            ':Advanced Option: Intended for advanced users and not used'
        )
        yield _indent(
            'by the majority of users, and might have a significant', 6
        )
        yield _indent('effect on stability and/or performance.', 6)

    if opt.sample_default:
        yield _indent('')
        yield _indent('This option has a sample default set, which means that')
        yield _indent(
            'its actual default value may vary from the one documented'
        )
        yield _indent('above.')

    try:
        help_text = opt.help % {'default': 'the value above'}
    except (TypeError, KeyError, ValueError):
        # There is no mention of the default in the help string,
        # the string had some unknown key, or the string contained
        # invalid formatting characters
        help_text = opt.help
    if help_text:
        yield ''
        for line in help_text.strip().splitlines():
            yield _indent(line.rstrip())

    # We don't bother outputting this if not using new-style choices with
    # inline descriptions
    if getattr(opt.type, 'choices', None) and not all(
        x is None for x in opt.type.choices.values()
    ):
        yield ''
        yield _indent('.. rubric:: Possible values')
        for choice in opt.type.choices:
            yield ''
            yield _indent(_get_choice_text(choice))
            yield _indent(
                _indent(
                    opt.type.choices[choice] or '<No description provided>'
                )
            )

    if opt.deprecated_opts:
        yield ''
        for line in _list_table(
            ['Group', 'Name'],
            (
                (d.group or group_name, d.name or opt.dest or 'UNSET')
                for d in opt.deprecated_opts
            ),
            title='Deprecated Variations',
        ):
            yield _indent(line)

    if opt.deprecated_for_removal:
        yield ''
        yield _indent('.. warning::')
        if opt.deprecated_since:
            yield _indent(
                '   This option is deprecated for removal '
                f'since {opt.deprecated_since}.'
            )
        else:
            yield _indent('   This option is deprecated for removal.')
        yield _indent('   Its value may be silently ignored ')
        yield _indent('   in the future.')
        if opt.deprecated_reason:
            reason = ' '.join(
                [x.strip() for x in opt.deprecated_reason.splitlines()]
            )
            yield ''
            yield _indent('   :Reason: ' + reason)

    yield ''


def _format_group(
    namespace: str | None,
    group_name: str,
    group_obj: cfg.OptGroup | None,
) -> Generator[str, None, None]:
    yield f'.. oslo.config:group:: {group_name}'
    if namespace:
        yield f'   :namespace: {namespace}'
    yield ''

    if group_obj and group_obj.help:
        for line in group_obj.help.strip().splitlines():
            yield _indent(line.rstrip())
        yield ''


def _format_group_opts(
    namespace: str | None,
    group_name: str | None,
    group_obj: cfg.OptGroup | None,
    opt_list: list[Any],
) -> Generator[str, None, None]:
    group_name = group_name or 'DEFAULT'
    LOG.debug('%s %s', namespace, group_name)

    yield from _format_group(namespace, group_name, group_obj)

    for opt in opt_list:
        yield from _format_opt(opt, group_name)


def _format_option_help(
    namespaces: list[str],
    split_namespaces: bool,
) -> Generator[str, None, None]:
    """Generate a series of lines of restructuredtext.

    Format the option help as restructuredtext and return it as a list
    of lines.
    """
    opts = generator._list_opts(namespaces)

    if split_namespaces:
        for namespace, opt_list in opts:
            for group, opts in opt_list:
                if isinstance(group, cfg.OptGroup):
                    group_name = group.name
                else:
                    group_name = group
                    group = None
                if group_name is None:
                    group_name = 'DEFAULT'
                lines = _format_group_opts(
                    namespace=namespace,
                    group_name=group_name,
                    group_obj=group,
                    opt_list=opts,
                )
                yield from lines
    else:
        # Merge the options from different namespaces that belong to
        # the same group together and format them without the
        # namespace.
        by_section: dict[str, list[Any]] = {}
        group_objs: dict[str, cfg.OptGroup | None] = {}
        for ignore, opt_list in opts:
            for group, group_opts in opt_list:
                if isinstance(group, cfg.OptGroup):
                    group_name = group.name
                else:
                    group_name = group
                    group = None
                if group_name is None:
                    group_name = 'DEFAULT'
                group_objs.setdefault(group_name, group)
                by_section.setdefault(group_name, []).extend(group_opts)
        for group_name, group_opts in sorted(by_section.items()):
            lines = _format_group_opts(
                namespace=None,
                group_name=group_name,
                group_obj=group_objs.get(group_name),
                opt_list=group_opts,
            )
            yield from lines


class ShowOptionsDirective(rst.Directive):
    option_spec = {
        'split-namespaces': directives.flag,
        'config-file': directives.unchanged,
    }

    has_content = True

    def run(self) -> list[nodes.Node]:
        split_namespaces = 'split-namespaces' in self.options

        config_file = self.options.get('config-file')
        if config_file:
            LOG.info('loading config file %s', config_file)
            conf = cfg.ConfigOpts()
            conf.register_opts(generator._generator_opts)
            conf(
                args=['--config-file', config_file],
                project='oslo.config.sphinxext',
            )
            namespaces = conf.namespace[:]
        else:
            namespaces = [c.strip() for c in self.content if c.strip()]

        result = StringList()
        source_name = self.state.document.current_source

        for count, line in enumerate(
            _format_option_help(namespaces, split_namespaces)
        ):
            result.append(line, source_name, count)
            LOG.debug('%5d%s%s', count, ' ' if line else '', line)

        node = nodes.section()
        node.document = self.state.document

        # With the resolution for bug #1755783, we now parse the 'Opt.help'
        # attribute as rST. Unfortunately, there are a lot of broken option
        # descriptions out there and we don't want to break peoples' builds
        # suddenly. As a result, we disable 'warning-is-error' temporarily.
        # Users will still see the warnings but the build will continue.
        with logging.skip_warningiserror():
            nested_parse_with_titles(self.state, result, node)

        return node.children


class ConfigGroupXRefRole(XRefRole):
    "Handles :oslo.config:group: roles pointing to configuration groups."

    def __init__(self) -> None:
        super().__init__(
            warn_dangling=True,
        )

    def process_link(
        self,
        env: BuildEnvironment,
        refnode: nodes.Element,
        has_explicit_title: bool,
        title: str,
        target: str,
    ) -> tuple[str, str]:
        # The anchor for the group link is the group name.
        return target, target


class ConfigOptXRefRole(XRefRole):
    "Handles :oslo.config:option: roles pointing to configuration options."

    def __init__(self) -> None:
        super().__init__(
            warn_dangling=True,
        )

    def process_link(
        self,
        env: BuildEnvironment,
        refnode: nodes.Element,
        has_explicit_title: bool,
        title: str,
        target: str,
    ) -> tuple[str, str]:
        if not has_explicit_title:
            title = target
        if '.' in target:
            group, opt_name = target.split('.')
        else:
            group = 'DEFAULT'
            opt_name = target
        anchor = _make_anchor_target(group, opt_name)
        return title, anchor


class ConfigGroup(rst.Directive):
    required_arguments = 1
    optional_arguments = 0
    has_content = True
    option_spec = {
        'namespace': directives.unchanged,
    }

    def run(self) -> list[nodes.Node]:
        env = self.state.document.settings.env

        group_name = self.arguments[0]
        namespace = self.options.get('namespace')

        cached_groups = env.domaindata['oslo.config']['groups']

        # Store the current group for use later in option directives
        env.temp_data['oslo.config:group'] = group_name
        LOG.debug('oslo.config group %s', group_name)

        # Store the location where this group is being defined
        # for use when resolving cross-references later.
        # FIXME: This should take the source namespace into account, too
        cached_groups[group_name] = env.docname

        result = StringList()
        source_name = '<' + __name__ + '>'

        def _add(text: str) -> None:
            "Append some text to the output result view to be parsed."
            result.append(text, source_name)

        if namespace:
            title = f'{namespace}: {group_name}'
        else:
            title = group_name

        _add(title)
        _add('-' * len(title))
        _add('')
        for line in self.content:
            _add(line)
        node = nodes.section()
        node.document = self.state.document
        nested_parse_with_titles(self.state, result, node)

        first_child = cast(nodes.Element, node.children[0])

        # Compute the normalized target and set the node to have that
        # as an id
        target_name = cfg._normalize_group_name(group_name)
        first_child['ids'].append(target_name)

        indexnode = addnodes.index(entries=[])
        return [indexnode] + node.children


class ConfigOption(ObjectDescription[str]):
    "Description of a configuration option (.. option)."

    def handle_signature(
        self, sig: str, signode: addnodes.desc_signature
    ) -> str:
        """Transform an option description into RST nodes."""
        optname = sig
        LOG.debug('oslo.config option %s', optname)
        # Insert a node into the output showing the option name
        signode += addnodes.desc_name(optname, optname)
        signode['allnames'] = [optname]
        return optname

    def add_target_and_index(
        self,
        firstname: str,
        sig: str,
        signode: addnodes.desc_signature,
    ) -> None:
        cached_options = self.env.domaindata['oslo.config']['options']
        # Look up the current group name from the processing context
        currgroup = self.env.temp_data.get('oslo.config:group') or 'DEFAULT'
        # Compute the normalized target name for the option and give
        # that to the node as an id
        target_name = _make_anchor_target(currgroup, sig)
        signode['ids'].append(target_name)
        self.state.document.note_explicit_target(signode)
        # Store the location of the option definition for later use in
        # resolving cross-references
        # FIXME: This should take the source namespace into account, too
        cached_options[target_name] = self.env.docname


class ConfigDomain(Domain):
    """oslo.config domain."""

    name = 'oslo.config'
    label = 'oslo.config'
    object_types = {
        'configoption': ObjType('configuration option', 'option'),
    }
    directives = {
        'group': ConfigGroup,
        'option': ConfigOption,
    }
    roles = {
        'option': ConfigOptXRefRole(),
        'group': ConfigGroupXRefRole(),
    }
    initial_data = {
        'options': {},
        'groups': {},
    }

    def resolve_xref(
        self,
        env: BuildEnvironment,
        fromdocname: str,
        builder: Builder,
        typ: str,
        target: str,
        node: addnodes.pending_xref,
        contnode: nodes.Element,
    ) -> nodes.reference | None:
        """Resolve cross-references"""
        if typ == 'option':
            group_name, option_name = target.split('.', 1)
            return make_refnode(
                builder,
                fromdocname,
                env.domaindata['oslo.config']['options'][target],
                target,
                contnode,
                option_name,
            )
        if typ == 'group':
            return make_refnode(
                builder,
                fromdocname,
                env.domaindata['oslo.config']['groups'][target],
                target,
                contnode,
                target,
            )
        return None

    def merge_domaindata(
        self, docnames: Set[str], otherdata: dict[str, Any]
    ) -> None:
        for target, docname in otherdata['options'].items():
            if docname in docnames:
                self.data['options'][target] = docname

        for target, docname in otherdata['groups'].items():
            if docname in docnames:
                self.data['groups'][target] = docname


def setup(app: Sphinx) -> dict[str, Any]:
    # NOTE(dhellmann): Try to turn off lazy translation from oslo_i18n
    # so any translated help text or deprecation messages associated
    # with configuration options are treated as regular strings
    # instead of Message objects. Unfortunately this is a bit
    # order-dependent, and so it's still possible that importing code
    # from another module such as through the autodoc features, or
    # even through the plugin scanner, will turn lazy evaluation back
    # on.
    oslo_i18n.enable_lazy(False)
    app.add_directive('show-options', ShowOptionsDirective)
    app.add_domain(ConfigDomain)
    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
r"""
Environment
-----------

The **environment** backend driver provides a method of accessing
configuration data in environment variables. It is enabled by default
and requires no additional configuration to use. The environment is
checked after command line options, but before configuration files.


Environment variables are checked for any configuration data. The variable
names take the form:

* A prefix of ``OS_``
* The group name, uppercased
* Separated from the option name by a `__` (double underscore)
* Followed by the name

For an option that looks like this in the usual INI format::

    [placement_database]
    connection = sqlite:///

the corresponding environment variable would be
``OS_PLACEMENT_DATABASE__CONNECTION``.

The Driver Class
================

.. autoclass:: EnvironmentConfigurationSourceDriver

The Configuration Source Class
==============================

.. autoclass:: EnvironmentConfigurationSource

"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

# Avoid circular import
import oslo_config.cfg
from oslo_config import sources

if TYPE_CHECKING:
    from oslo_config import cfg


# In current practice this class is not used because the
# EnvironmentConfigurationSource is loaded by default, but we keep it
# here in case we choose to change that behavior in the future.
class EnvironmentConfigurationSourceDriver(sources.ConfigurationSourceDriver):
    """A backend driver for environment variables.

    This configuration source is available by default and does not need special
    configuration to use. The sample config is generated automatically but is
    not necessary.
    """

    def list_options_for_discovery(self) -> list[Any]:
        """There are no options for this driver."""
        return []

    def open_source_from_opt_group(
        self,
        conf: cfg.ConfigOpts,
        group_name: str,
    ) -> EnvironmentConfigurationSource:
        return EnvironmentConfigurationSource()


class EnvironmentConfigurationSource(sources.ConfigurationSource):
    """A configuration source for options in the environment."""

    @staticmethod
    def get_name(group_name: str | None, option_name: str) -> str:
        """Return the expected environment variable name for the given option.

        :param group_name: The group name or None. Defaults to 'DEFAULT' if
            None.
        :param option_name: The option name.
        :returns: Th expected environment variable name.
        """
        group_name = group_name or 'DEFAULT'
        return f'OS_{group_name.upper()}__{option_name.upper()}'

    def get(
        self,
        group_name: str | None,
        option_name: str,
        opt: cfg.Opt,
    ) -> tuple[Any, cfg.LocationInfo | None]:
        env_name = self.get_name(group_name, option_name)
        try:
            value = os.environ[env_name]
            loc = oslo_config.cfg.LocationInfo(
                oslo_config.cfg.Locations.environment, env_name
            )
            return (value, loc)
        except KeyError:
            return (sources._NoValue, None)



# Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections.abc import Iterable
from typing import NoReturn


class ParseError(Exception):
    def __init__(self, message: str, lineno: int, line: str) -> None:
        self.msg = message
        self.line = line
        self.lineno = lineno

    def __str__(self) -> str:
        return f'at line {self.lineno}, {self.msg}: {self.line!r}'


class BaseParser:
    lineno: int = 0
    parse_exc: type[ParseError] = ParseError

    def _assignment(
        self, key: str, value: list[str]
    ) -> tuple[None, list[str]]:
        self.assignment(key, value)
        return None, []

    def _get_section(self, line: str) -> str:
        if not line.endswith(']'):
            return self.error_no_section_end_bracket(line)
        if len(line) <= 2:
            return self.error_no_section_name(line)

        return line[1:-1]

    def _split_key_value(self, line: str) -> tuple[str, list[str]]:
        colon = line.find(':')
        equal = line.find('=')
        if colon < 0 and equal < 0:
            return self.error_invalid_assignment(line)

        if colon < 0 or (equal >= 0 and equal < colon):
            key, value = line[:equal], line[equal + 1 :]
        else:
            key, value = line[:colon], line[colon + 1 :]

        value = value.strip()
        if value and value[0] == value[-1] and value.startswith(("\"", "'")):
            value = value[1:-1]
        return key.strip(), [value]

    def parse(self, lineiter: Iterable[str]) -> None:
        key: str | None = None
        value: list[str] = []

        for line in lineiter:
            self.lineno += 1

            line = line.rstrip()
            if not line:
                # Blank line, ends multi-line values
                if key:
                    key, value = self._assignment(key, value)
                continue
            elif line.startswith((' ', '\t')):
                # Continuation of previous assignment
                if key is None:
                    self.error_unexpected_continuation(line)
                else:
                    value.append(line.lstrip())
                continue

            if key:
                # Flush previous assignment, if any
                key, value = self._assignment(key, value)

            if line.startswith('['):
                # Section start
                section = self._get_section(line)
                if section:
                    self.new_section(section)
            elif line.startswith(('#', ';')):
                self.comment(line[1:].lstrip())
            else:
                key, value = self._split_key_value(line)
                if not key:
                    self.error_empty_key(line)

        if key:
            # Flush previous assignment, if any
            self._assignment(key, value)

    def assignment(self, key: str, value: list[str]) -> None:
        """Called when a full assignment is parsed."""
        raise NotImplementedError()

    def new_section(self, section: str) -> None:
        """Called when a new section is started."""
        raise NotImplementedError()

    def comment(self, comment: str) -> None:
        """Called when a comment is parsed."""
        pass

    def error_invalid_assignment(self, line: str) -> NoReturn:
        raise self.parse_exc(
            "No ':' or '=' found in assignment", self.lineno, line
        )

    def error_empty_key(self, line: str) -> NoReturn:
        raise self.parse_exc('Key cannot be empty', self.lineno, line)

    def error_unexpected_continuation(self, line: str) -> NoReturn:
        raise self.parse_exc('Unexpected continuation line', self.lineno, line)

    def error_no_section_end_bracket(self, line: str) -> NoReturn:
        raise self.parse_exc(
            'Invalid section (must end with ])', self.lineno, line
        )

    def error_no_section_name(self, line: str) -> NoReturn:
        raise self.parse_exc('Empty section name', self.lineno, line)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslotest import base

from oslo_config import sphinxconfiggen


class SingleSampleGenerationTest(base.BaseTestCase):
    @mock.patch('os.path.isdir')
    @mock.patch('os.path.isfile')
    @mock.patch('oslo_config.generator.main')
    def test_sample_gen_with_single_config_file(self, main, isfile, isdir):
        isfile.side_effect = [False, True]
        isdir.return_value = True

        config = mock.Mock(
            config_generator_config_file='nova-gen.conf',
            sample_config_basename='nova',
        )
        app = mock.Mock(srcdir='/opt/nova', config=config)
        sphinxconfiggen.generate_sample(app)

        main.assert_called_once_with(
            args=[
                '--config-file',
                '/opt/nova/nova-gen.conf',
                '--output-file',
                '/opt/nova/nova.conf.sample',
            ]
        )

    @mock.patch('os.path.isdir')
    @mock.patch('os.path.isfile')
    @mock.patch('oslo_config.generator.main')
    def test_sample_gen_with_single_config_file_no_base(
        self, main, isfile, isdir
    ):
        isfile.side_effect = [False, True]
        isdir.return_value = True

        config = mock.Mock(
            config_generator_config_file='nova-gen.conf',
            sample_config_basename=None,
        )
        app = mock.Mock(srcdir='/opt/nova', config=config)
        sphinxconfiggen.generate_sample(app)

        main.assert_called_once_with(
            args=[
                '--config-file',
                '/opt/nova/nova-gen.conf',
                '--output-file',
                '/opt/nova/sample.config',
            ]
        )


class MultipleSampleGenerationTest(base.BaseTestCase):
    @mock.patch('os.path.isdir')
    @mock.patch('os.path.isfile')
    @mock.patch('oslo_config.generator.main')
    def test_multi_sample_gen(self, main, isfile, isdir):
        isfile.side_effect = [False, True, False, True]
        isdir.return_value = True

        multiple_configs = [
            ('glance-api-gen.conf', 'glance-api'),
            ('glance-reg-gen.conf', 'glance-reg'),
        ]
        config = mock.Mock(config_generator_config_file=multiple_configs)
        app = mock.Mock(srcdir='/opt/glance', config=config)
        sphinxconfiggen.generate_sample(app)

        self.assertEqual(main.call_count, 2)
        main.assert_any_call(
            args=[
                '--config-file',
                '/opt/glance/glance-api-gen.conf',
                '--output-file',
                '/opt/glance/glance-api.conf.sample',
            ]
        )
        main.assert_any_call(
            args=[
                '--config-file',
                '/opt/glance/glance-reg-gen.conf',
                '--output-file',
                '/opt/glance/glance-reg.conf.sample',
            ]
        )

    @mock.patch('os.path.isdir')
    @mock.patch('os.path.isfile')
    @mock.patch('oslo_config.generator.main')
    def test_multi_sample_gen_with_without_one_base(self, main, isfile, isdir):
        isfile.side_effect = [False, True, False, True]
        isdir.return_value = True

        multiple_configs = [
            ('glance-api-gen.conf', 'glance-api'),
            ('glance-reg-gen.conf', None),
        ]
        config = mock.Mock(config_generator_config_file=multiple_configs)
        app = mock.Mock(srcdir='/opt/glance', config=config)
        sphinxconfiggen.generate_sample(app)

        self.assertEqual(main.call_count, 2)
        main.assert_any_call(
            args=[
                '--config-file',
                '/opt/glance/glance-api-gen.conf',
                '--output-file',
                '/opt/glance/glance-api.conf.sample',
            ]
        )
        main.assert_any_call(
            args=[
                '--config-file',
                '/opt/glance/glance-reg-gen.conf',
                '--output-file',
                '/opt/glance/glance-reg-gen.conf.sample',
            ]
        )

    @mock.patch('os.path.isdir')
    @mock.patch('os.path.isfile')
    @mock.patch('oslo_config.generator.main')
    def test_multi_sample_gen_with_without_any_base(self, main, isfile, isdir):
        isfile.side_effect = [False, True, False, True]
        isdir.return_value = True

        multiple_configs = [
            ('glance-api-gen.conf', None),
            ('glance-reg-gen.conf', None),
        ]
        config = mock.Mock(config_generator_config_file=multiple_configs)
        app = mock.Mock(srcdir='/opt/glance', config=config)
        sphinxconfiggen.generate_sample(app)

        self.assertEqual(main.call_count, 2)
        main.assert_any_call(
            args=[
                '--config-file',
                '/opt/glance/glance-api-gen.conf',
                '--output-file',
                '/opt/glance/glance-api-gen.conf.sample',
            ]
        )
        main.assert_any_call(
            args=[
                '--config-file',
                '/opt/glance/glance-reg-gen.conf',
                '--output-file',
                '/opt/glance/glance-reg-gen.conf.sample',
            ]
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
r"""
Remote File
-----------

The **remote_file** backend driver is the first driver implemented by
oslo.config. It extends the previous limit of only accessing local files
to a new scenario where it is possible to access configuration data over
the network. The **remote_file** driver is based on the **requests** module
and is capable of accessing remote files through **HTTP** or **HTTPS**.

To definition of a remote_file configuration data source can be as minimal as::

    [DEFAULT]
    config_source = external_config_group

    [external_config_group]
    driver = remote_file
    uri = http://mydomain.com/path/to/config/data.conf

Or as complete as::

    [DEFAULT]
    config_source = external_config_group

    [external_config_group]
    driver = remote_file
    uri = https://mydomain.com/path/to/config/data.conf
    ca_path = /path/to/server/ca.pem
    client_key = /path/to/my/key.pem
    client_cert = /path/to/my/cert.pem

On the following sessions, you can find more information about this driver's
classes and its options.

The Driver Class
================

.. autoclass:: URIConfigurationSourceDriver

The Configuration Source Class
==============================

.. autoclass:: URIConfigurationSource

"""

from typing import Any

import requests
import tempfile

from oslo_config import cfg
from oslo_config import sources


class URIConfigurationSourceDriver(sources.ConfigurationSourceDriver):
    """A backend driver for remote files served through http[s].

    Required options:
      - uri: URI containing the file location.

    Non-required options:
      - ca_path: The path to a CA_BUNDLE file or directory with
                 certificates of trusted CAs.

      - client_cert: Client side certificate, as a single file path
                     containing either the certificate only or the
                     private key and the certificate.

      - client_key: Client side private key, in case client_cert is
                    specified but does not includes the private key.
    """

    _uri_driver_opts = [
        cfg.URIOpt(
            'uri',
            schemes=['http', 'https'],
            required=True,
            sample_default='https://example.com/my-configuration.ini',
            help=(
                'Required option with the URI of the '
                'extra configuration file\'s location.'
            ),
        ),
        cfg.StrOpt(
            'ca_path',
            sample_default='/etc/ca-certificates',
            help=(
                'The path to a CA_BUNDLE file or directory '
                'with certificates of trusted CAs.'
            ),
        ),
        cfg.StrOpt(
            'client_cert',
            sample_default='/etc/ca-certificates/service-client-keystore',
            help=(
                'Client side certificate, as a single file path '
                'containing either the certificate only or the '
                'private key and the certificate.'
            ),
        ),
        cfg.StrOpt(
            'client_key',
            help=(
                'Client side private key, in case client_cert is '
                'specified but does not includes the private key.'
            ),
        ),
        cfg.IntOpt(
            'timeout',
            default=60,
            help=(
                'Timeout is the number of seconds the request will wait '
                'for your client to establish a connection to a remote '
                'machine call on the socket.'
            ),
        ),
    ]

    def list_options_for_discovery(self) -> list[Any]:
        return self._uri_driver_opts

    def open_source_from_opt_group(
        self,
        conf: cfg.ConfigOpts,
        group_name: str,
    ) -> 'URIConfigurationSource':
        conf.register_opts(self._uri_driver_opts, group_name)

        return URIConfigurationSource(
            conf[group_name].uri,
            conf[group_name].ca_path,
            conf[group_name].client_cert,
            conf[group_name].client_key,
            conf[group_name].timeout,
        )


class URIConfigurationSource(sources.ConfigurationSource):
    """A configuration source for remote files served through http[s].

    :param uri: The Uniform Resource Identifier of the configuration to be
          retrieved.

    :param ca_path: The path to a CA_BUNDLE file or directory with
              certificates of trusted CAs.

    :param client_cert: Client side certificate, as a single file path
                  containing either the certificate only or the
                  private key and the certificate.

    :param client_key: Client side private key, in case client_cert is
                 specified but does not includes the private key.
    """

    def __init__(
        self,
        uri: str,
        ca_path: str | None = None,
        client_cert: str | None = None,
        client_key: str | None = None,
        timeout: int = 60,
    ) -> None:
        self._uri = uri
        self._namespace = cfg._Namespace(cfg.ConfigOpts())

        data = self._fetch_uri(uri, ca_path, client_cert, client_key, timeout)

        with tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.write(data.encode("utf-8"))
            tmpfile.flush()

            cfg.ConfigParser._parse_file(tmpfile.name, self._namespace)

    def _fetch_uri(
        self,
        uri: str,
        ca_path: str | None,
        client_cert: str | None,
        client_key: str | None,
        timeout: int,
    ) -> str:
        verify: str | bool = ca_path if ca_path else True
        cert: tuple[str, str] | str | None = (
            (client_cert, client_key)
            if client_cert and client_key
            else client_cert
        )

        with requests.get(
            uri, verify=verify, cert=cert, timeout=timeout
        ) as response:
            response.raise_for_status()  # raises only in case of HTTPError
            return response.text

    def get(
        self,
        group_name: str | None,
        option_name: str,
        opt: cfg.Opt,
    ) -> tuple[Any, cfg.LocationInfo | None]:
        try:
            result: tuple[Any, cfg.LocationInfo | None] = (
                self._namespace._get_value(
                    [(group_name, option_name)], multi=opt.multi
                )
            )
            return result
        except KeyError:
            return (sources._NoValue, None)
